@@ -55,16 +55,21 @@ struct _GSSDPResourceGroupPrivate {
 
         gulong       message_received_id;
 
-        guint        timeout_id;
+        GSource     *timeout_src;
 
         guint        last_resource_id;
+        
+        guint        message_delay;
+        GQueue      *message_queue;
+        guint        message_src_id;
 };
 
 enum {
         PROP_0,
         PROP_CLIENT,
         PROP_MAX_AGE,
-        PROP_AVAILABLE
+        PROP_AVAILABLE,
+        PROP_MESSAGE_DELAY
 };
 
 typedef struct {
@@ -77,6 +82,8 @@ typedef struct {
         GList               *responses;
 
         guint                id;
+
+        gboolean             initial_alive_sent;
 } Resource;
 
 typedef struct {
@@ -85,9 +92,11 @@ typedef struct {
         char     *target;
         Resource *resource;
 
-        guint     timeout_id;
+        GSource  *timeout_src;
 } DiscoveryResponse;
 
+#define DEFAULT_MESSAGE_DELAY 20 
+
 /* Function prototypes */
 static void
 gssdp_resource_group_set_client (GSSDPResourceGroup *resource_group,
@@ -111,6 +120,8 @@ static gboolean
 discovery_response_timeout      (gpointer            user_data);
 static void
 discovery_response_free         (DiscoveryResponse  *response);
+static gboolean
+process_queue                   (gpointer            data);
 
 static void
 gssdp_resource_group_init (GSSDPResourceGroup *resource_group)
@@ -121,6 +132,9 @@ gssdp_resource_group_init (GSSDPResourceGroup *resource_group)
                                          GSSDPResourceGroupPrivate);
 
         resource_group->priv->max_age = SSDP_DEFAULT_MAX_AGE;
+        resource_group->priv->message_delay = DEFAULT_MESSAGE_DELAY;
+
+        resource_group->priv->message_queue = g_queue_new ();
 }
 
 static void
@@ -149,6 +163,12 @@ gssdp_resource_group_get_property (GObject    *object,
                         (value,
                          gssdp_resource_group_get_available (resource_group));
                 break;
+        case PROP_MESSAGE_DELAY:
+                g_value_set_uint
+                        (value,
+                         gssdp_resource_group_get_message_delay 
+                                (resource_group));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -178,6 +198,10 @@ gssdp_resource_group_set_property (GObject      *object,
                 gssdp_resource_group_set_available
                         (resource_group, g_value_get_boolean (value));
                 break;
+        case PROP_MESSAGE_DELAY:
+                gssdp_resource_group_set_message_delay
+                        (resource_group, g_value_get_uint (value));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -188,32 +212,53 @@ static void
 gssdp_resource_group_dispose (GObject *object)
 {
         GSSDPResourceGroup *resource_group;
+        GSSDPResourceGroupPrivate *priv;
 
         resource_group = GSSDP_RESOURCE_GROUP (object);
+        priv = resource_group->priv;
 
-        while (resource_group->priv->resources) {
-                resource_free (resource_group->priv->resources->data);
-                resource_group->priv->resources =
-                        g_list_delete_link (resource_group->priv->resources,
-                                            resource_group->priv->resources);
+        while (priv->resources) {
+                resource_free (priv->resources->data);
+                priv->resources =
+                        g_list_delete_link (priv->resources,
+                                            priv->resources);
         }
 
-        if (resource_group->priv->timeout_id > 0) {
-                g_source_remove (resource_group->priv->timeout_id);
-                resource_group->priv->timeout_id = 0;
+        if (priv->message_queue) {
+                /* send messages without usual delay */
+                while (!g_queue_is_empty (priv->message_queue)) {
+                        if (priv->available)
+                                process_queue (resource_group);
+                        else
+                                g_free (g_queue_pop_head
+                                        (priv->message_queue));
+                }
+
+                g_queue_free (priv->message_queue);
+                priv->message_queue = NULL;
         }
 
-        if (resource_group->priv->client) {
+        if (priv->message_src_id > 0) {
+                g_source_remove (priv->message_src_id);
+                priv->message_src_id = 0;
+        }
+
+        if (priv->timeout_src) {
+                g_source_destroy (priv->timeout_src);
+                priv->timeout_src = NULL;
+        }
+
+        if (priv->client) {
                 if (g_signal_handler_is_connected
-                        (resource_group->priv->client,
-                         resource_group->priv->message_received_id)) {
+                        (priv->client,
+                         priv->message_received_id)) {
                         g_signal_handler_disconnect
-                                (resource_group->priv->client,
-                                 resource_group->priv->message_received_id);
+                                (priv->client,
+                                 priv->message_received_id);
                 }
                                                    
-                g_object_unref (resource_group->priv->client);
-                resource_group->priv->client = NULL;
+                g_object_unref (priv->client);
+                priv->client = NULL;
         }
 }
 
@@ -283,6 +328,27 @@ gssdp_resource_group_class_init (GSSDPResourceGroupClass *klass)
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                           G_PARAM_STATIC_BLURB));
+
+        /**
+         * GSSDPResourceGroup:message-delay
+         *
+         * The minimum number of milliseconds between SSDP messages.
+         * The default is 20 based on DLNA specification.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_MESSAGE_DELAY,
+                 g_param_spec_uint
+                         ("message-delay",
+                          "Message delay",
+                          "The minimum number of milliseconds between SSDP "
+                          "messages.",
+                          0,
+                          G_MAXUINT,
+                          DEFAULT_MESSAGE_DELAY,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
+                          G_PARAM_STATIC_BLURB));
 }
 
 /**
@@ -338,7 +404,7 @@ gssdp_resource_group_get_client (GSSDPResourceGroup *resource_group)
 /**
  * gssdp_resource_group_set_max_age
  * @resource_group: A #GSSDPResourceGroup
- * @mx: The number of seconds advertisements are valid
+ * @max_age: The number of seconds advertisements are valid
  *
  * Sets the number of seconds advertisements are valid to @max_age.
  **/
@@ -370,6 +436,41 @@ gssdp_resource_group_get_max_age (GSSDPResourceGroup *resource_group)
         return resource_group->priv->max_age;
 }
 
+/**
+ * gssdp_resource_group_set_message_delay
+ * @resource_group: A #GSSDPResourceGroup
+ * @message_delay: The message delay in ms.
+ *
+ * Sets the minimum time between each SSDP message.
+ **/
+void
+gssdp_resource_group_set_message_delay (GSSDPResourceGroup *resource_group,
+                                        guint               message_delay)
+{
+        g_return_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group));
+
+        if (resource_group->priv->message_delay == message_delay)
+                return;
+
+        resource_group->priv->message_delay = message_delay;
+        
+        g_object_notify (G_OBJECT (resource_group), "message-delay");
+}
+
+/**
+ * gssdp_resource_group_get_message_delay
+ * @resource_group: A #GSSDPResourceGroup
+ *
+ * Return value: the minimum time between each SSDP message in ms.
+ **/
+guint
+gssdp_resource_group_get_message_delay (GSSDPResourceGroup *resource_group)
+{
+        g_return_val_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group), 0);
+
+        return resource_group->priv->message_delay;
+}
+
 /**
  * gssdp_resource_group_set_available
  * @resource_group: A #GSSDPResourceGroup
@@ -393,11 +494,26 @@ gssdp_resource_group_set_available (GSSDPResourceGroup *resource_group,
         resource_group->priv->available = available;
 
         if (available) {
+                GMainContext *context;
+                int timeout;
+
+                /* We want to re-announce before actually timing out */
+                timeout = resource_group->priv->max_age;
+                if (timeout > 2)
+                        timeout = timeout / 2 - 1;
+
                 /* Add re-announcement timer */
-                resource_group->priv->timeout_id =
-                        g_timeout_add_seconds (resource_group->priv->max_age,
-                                               resource_group_timeout,
-                                               resource_group);
+                resource_group->priv->timeout_src =
+                        g_timeout_source_new_seconds (timeout);
+		g_source_set_callback (resource_group->priv->timeout_src,
+				       resource_group_timeout,
+				       resource_group, NULL);
+
+                context = gssdp_client_get_main_context
+                        (resource_group->priv->client);
+		g_source_attach (resource_group->priv->timeout_src, context);
+
+                g_source_unref (resource_group->priv->timeout_src);
 
                 /* Announce all resources */
                 for (l = resource_group->priv->resources; l; l = l->next)
@@ -408,8 +524,8 @@ gssdp_resource_group_set_available (GSSDPResourceGroup *resource_group,
                         resource_byebye (l->data);
 
                 /* Remove re-announcement timer */
-                g_source_remove (resource_group->priv->timeout_id);
-                resource_group->priv->timeout_id = 0;
+                g_source_destroy (resource_group->priv->timeout_src);
+                resource_group->priv->timeout_src = NULL;
         }
         
         g_object_notify (G_OBJECT (resource_group), "available");
@@ -462,6 +578,8 @@ gssdp_resource_group_add_resource (GSSDPResourceGroup *resource_group,
         resource->target = g_strdup (target);
         resource->usn    = g_strdup (usn);
 
+        resource->initial_alive_sent = FALSE;
+
         for (l = locations; l; l = l->next) {
                 resource->locations = g_list_append (resource->locations,
                                                     g_strdup (l->data));
@@ -558,7 +676,7 @@ gssdp_resource_group_remove_resource (GSSDPResourceGroup *resource_group,
 }
 
 /**
- * Called every max-age seconds to re-announce all resources
+ * Called to re-announce all resources periodically
  **/
 static gboolean
 resource_group_timeout (gpointer user_data)
@@ -630,6 +748,7 @@ message_received_cb (GSSDPClient        *client,
                         /* Match. */
                         guint timeout;
                         DiscoveryResponse *response;
+                        GMainContext *context;
 
                         /* Get a random timeout from the interval [0, mx] */
                         timeout = g_random_int_range (0, mx * 1000);
@@ -639,14 +758,23 @@ message_received_cb (GSSDPClient        *client,
                         
                         response->dest_ip   = g_strdup (from_ip);
                         response->dest_port = from_port;
-                        response->target    = g_strdup (target);
                         response->resource  = resource;
 
+                        if (want_all)
+                                response->target = g_strdup (resource->target);
+                        else
+                                response->target = g_strdup (target);
+
                         /* Add timeout */
-                        response->timeout_id =
-                                g_timeout_add (timeout,
-                                               discovery_response_timeout,
-                                               response);
+                        response->timeout_src = g_timeout_source_new (timeout);
+			g_source_set_callback (response->timeout_src,
+					       discovery_response_timeout,
+					       response, NULL);
+
+                        context = gssdp_client_get_main_context (client);
+			g_source_attach (response->timeout_src, context);
+
+                        g_source_unref (response->timeout_src);
                         
                         /* Add to resource */
                         resource->responses =
@@ -737,7 +865,7 @@ discovery_response_free (DiscoveryResponse *response)
         response->resource->responses =
                 g_list_remove (response->resource->responses, response);
 
-        g_source_remove (response->timeout_id);
+        g_source_destroy (response->timeout_src);
         
         g_free (response->dest_ip);
         g_free (response->target);
@@ -745,6 +873,63 @@ discovery_response_free (DiscoveryResponse *response)
         g_slice_free (DiscoveryResponse, response);
 }
 
+/**
+ * Send the next queued message, if any
+ **/
+static gboolean
+process_queue (gpointer data)
+{
+        GSSDPResourceGroup *resource_group;
+
+        resource_group = GSSDP_RESOURCE_GROUP (data);
+
+        if (g_queue_is_empty (resource_group->priv->message_queue)) {
+                /* this is the timeout after last message in queue */
+                resource_group->priv->message_src_id = 0;
+
+                return FALSE;
+        } else {
+                GSSDPClient *client;
+                char *message;
+
+                client = resource_group->priv->client;
+                message = g_queue_pop_head
+                        (resource_group->priv->message_queue);
+
+                _gssdp_client_send_message (client,
+                                            NULL,
+                                            0,
+                                            message);
+                g_free (message);
+
+                return TRUE;
+        }
+}
+
+/**
+ * Add a message to sending queue
+ * 
+ * Do not free @message.
+ **/
+static void
+queue_message (GSSDPResourceGroup *resource_group,
+               char               *message)
+{
+        g_queue_push_tail (resource_group->priv->message_queue, 
+                           message);
+
+        if (resource_group->priv->message_src_id == 0) {
+                /* nothing in the queue: process message immediately 
+                   and add a timeout for (possible) next message */
+
+                process_queue (resource_group);
+                resource_group->priv->message_src_id = 
+                        g_timeout_add (resource_group->priv->message_delay, 
+                                       process_queue,
+                                       resource_group);
+        }
+}
+
 /**
  * Send ssdp:alive message for @resource
  **/
@@ -755,6 +940,15 @@ resource_alive (Resource *resource)
         guint max_age;
         char *al, *message;
 
+        if (!resource->initial_alive_sent) {
+                /* Unannounce before first announce. This is done to
+                   minimize the possibility of control points thinking
+                   that this is just a reannouncement. */
+                resource_byebye (resource);
+
+                resource->initial_alive_sent = TRUE;
+        }
+
         /* Send message */
         client = resource->resource_group->priv->client;
 
@@ -770,12 +964,8 @@ resource_alive (Resource *resource)
                                    resource->target,
                                    resource->usn);
 
-        _gssdp_client_send_message (client,
-                                    NULL,
-                                    0,
-                                    message);
+        queue_message (resource->resource_group, message);
 
-        g_free (message);
         g_free (al);
 }
 
@@ -785,22 +975,14 @@ resource_alive (Resource *resource)
 static void
 resource_byebye (Resource *resource)
 {
-        GSSDPClient *client;
         char *message;
 
-        /* Send message */
-        client = resource->resource_group->priv->client;
-
+        /* Queue message */
         message = g_strdup_printf (SSDP_BYEBYE_MESSAGE,
                                    resource->target,
                                    resource->usn);
         
-        _gssdp_client_send_message (client,
-                                    NULL,
-                                    0,
-                                    message);
-
-        g_free (message);
+        queue_message (resource->resource_group, message);
 }
 
 /**
