@@ -1,5 +1,5 @@
 /* 
- * Copyright (C) 2007 Zeeshan Ali <zeenix@gstreamer.net>
+ * Copyright (C) 2007 Zeeshan Ali (Khattak) <zeeshanak@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -596,14 +596,15 @@ init_upnp (void)
 {
         GError *error;
         
-        g_thread_init (NULL);
-
         error = NULL;
         client = gssdp_client_new (NULL, &error);
         if (error) {
-                g_critical (error->message);
+                g_printerr ("Error creating the GSSDP client: %s\n",
+                            error->message);
+
                 g_error_free (error);
-                return 1;
+
+                return FALSE;
         }
 
         resource_browser = gssdp_resource_browser_new (client,
